#!/bin/bash
set -e -o pipefail

readonly BLUE_GREEN_STRATEGY="blueGreen"
readonly SCRIPT_CONTEXT_BUILD_NODE="buildNode"
readonly SCRIPT_CONTEXT_TARGET_CLUSTER="targetCluster"
readonly SCRIPT_CONTEXT_CURRENT_CLUSTER="currentCluster"
readonly DEFAULT_SCRIPT_NAME="anonymous"

downloadBuildInfo() {
  # download buildInfo artifacts to tardir
  local buildinfo_res_name=$1
  local integration_alias=$2
  local buildinfo_name=$(eval echo "$"res_"$buildinfo_res_name"_buildName)
  local buildinfo_number=$(eval echo "$"res_"$buildinfo_res_name"_buildNumber)

  if [ -z "$buildinfo_name" ] || [ -z "$buildinfo_number" ]; then
    execute_command "echo buildName and buildNumber are required to use a BuildInfo resource with this step."
    exit 1
  fi

  local rt_url=$(eval echo "$"res_"$buildinfo_res_name"_"$integration_alias"_url)
  local rt_user=$(eval echo "$"res_"$buildinfo_res_name"_"$integration_alias"_user)
  local rt_apikey=$(eval echo "$"res_"$buildinfo_res_name"_"$integration_alias"_apikey)
  execute_command "retry_command $jfrog_cli_path rt download \"*\" $tardir/ --build=$buildinfo_name/$buildinfo_number --url=$rt_url --user=$rt_user --password=\$rt_apikey --insecure-tls=$no_verify_ssl"
}

downloadReleaseBundle() {
  # download releaseBundle artifacts to tardir
  local rt_int_name=$1
  local releaseBundle_res_name=$2
  local rt_url=$(eval echo "$"int_"$rt_int_name"_url)
  local rt_user=$(eval echo "$"int_"$rt_int_name"_user)
  local rt_apikey=$(eval echo "$"int_"$rt_int_name"_apikey)
  local releaseBundle_name=$(eval echo "$"res_"$releaseBundle_res_name"_name)
  local releaseBundle_version=$(eval echo "$"res_"$releaseBundle_res_name"_version)
  execute_command "retry_command $jfrog_cli_path rt download \"*\" $tardir/ --bundle=$releaseBundle_name/$releaseBundle_version --url=$rt_url --user=$rt_user --password=\$rt_apikey --insecure-tls=$no_verify_ssl"
}

# 'execute_on' functions are documented externally and available to users of LinuxVMDeploy
# for use in lifeCycle handlers like onComplete and onFailure.

execute_on_build_node() {
  if [[ $# -le 1 ]]; then
    echo "Usage: execute_on_build_node COMMAND_NAME COMMAND" >&2
    exit 99
  fi

  local command_name="$1"
  local command="$2"
  if [ "$dry_run" == true ]; then
    command="echo \"dry_run is true. $command_name will not run\""
  fi
  echo "Running command: '$command_name' | On build node."
  printf "%s\n%s\n" "Name: $command_name" "Command: \"$command\""
  eval "$command"
}

execute_on_target_cluster() {
  if [[ $# -le 1 ]]; then
    echo "Usage: execute_on_target_cluster COMMAND_NAME COMMAND" >&2
    exit 99
  fi

  local command_name="$1"
  local command="$2"
  execute_on_cluster "$target_cluster_name" "$command_name" "$command"
}

execute_on_current_cluster() {
  if [[ $# -le 1 ]]; then
    echo "Usage: execute_on_current_cluster COMMAND_NAME COMMAND" >&2
    exit 99
  fi

  local command_name="$1"
  local command="$2"
  execute_on_cluster "$current_cluster_name" "$command_name" "$command"
}

execute_on_cluster() {
  if [[ $# -le 2 ]]; then
    echo "Usage: execute_on_cluster VM_CLUSTER_RESOURCE_NAME COMMAND_NAME COMMAND" >&2
    exit 99
  fi

  local cluster="$1"
  local command_name="$2"
  local command="$3"
  local res_targets=$(eval echo "$"res_"$cluster"_targets)
  local vm_targets=( $(echo "$res_targets" | jq --raw-output '.[]') )

  if [ -n "$deploy_targets_override" ] && [ "$strategy" != "$BLUE_GREEN_STRATEGY" ]; then
    echo "Overriding vm deploy targets with: $deploy_targets_override"
    IFS=,
    vm_targets=($deploy_targets_override)
    unset IFS
  fi

  execute_command "echo \"Running command: '$command_name' | On cluster: '$cluster'\""

  for i in $(seq 0 $(("${#vm_targets[@]}"-1)))
  do
    local vm_target=$(eval echo "${vm_targets[$i]}")

    # Wait between deploys if delay was specified
    if [ -n "$rollout_delay_config_val" ] && [ "$i" != 0 ]; then
      execute_command "sleep ${rollout_delay_config_val}s"
    fi
    local ssh_base_command="ssh $ssh_user_config_val@$vm_target -o StrictHostKeyChecking=no"

    # Command to source the file with vmEnvironmentVariables
    local source_env_file
    # There is only an env file on the target_cluster.
    if [ -n "$vm_env_vars_len_config_val" ] && [ "$cluster" == "$target_cluster_name" ]; then
      source_env_file="source $target_dir/$vm_env_filename;"
    fi

    if [ "$dry_run" == true ]; then
      # Echo a message on the vm instead of running the command.
      command="echo \"dry_run is true. $command_name will not run\""
    fi

    # Command to run the provided command from within the uploaded dir with env vars set
    local full_command="$ssh_base_command \"$source_env_file \$command\""

    execute_command "echo \"Running command: '$command_name' | On vm: '$vm_target'\""
    echo "--- Command ---"
    printf "%s\n%s\n%s\n" "Name: $command_name" "VM Address: $vm_target" "Command: \"$command\""
    echo "--- Command Output ---"
    eval "$full_command" || eval "$on_failure"

    # rollback archives will be created on deployed_vms with rollingUpdate strategy.
    # It does not matter for blueGreen.
    deployed_vms+=("$vm_target");
  done
  exitIfDeployFailed
}

exitIfDeployFailed() {
  if [ "$deploy_failed" == true ]; then
    # There was a failure, so we exit and do not create a new rollback archive
    execute_command "echo 'Exiting due to failed deployment.'"
    execute_command "exit 1"
  fi
}

DeployApplication() {
  # export current_command for use in onFailure.
  export current_command=""

  local strategy=$(find_step_configuration_value 'strategy')
  if [ -z "$strategy" ]; then
    execute_command "echo 'Defaulting to rollingUpdate strategy.'"
    strategy="rollingUpdate"
  fi

  local vm_cluster_names=$(get_resource_names --type VmCluster --operation IN)

  if [ -z "$vm_cluster_names" ]; then
    execute_command "echo 'A VmCluster input resource is required.'"
    execute_command "exit 1"
  fi

  local vm_cluster_count=$(echo "$vm_cluster_names" | jq length)

  # validate config related to blueGreen strategy
  if [ "$strategy" == "$BLUE_GREEN_STRATEGY" ] && [ "$vm_cluster_count" -ne 2 ]; then
    execute_command "echo 'Exactly two VmCluster resources are required when using blueGreen strategy.'"
    execute_command "exit 1"
  fi

  local buildinfo_res_name=$(get_resource_name --type BuildInfo --operation IN)
  local filespec_res_name=$(get_resource_name --type FileSpec --operation IN)
  local releaseBundle_res_name=$(get_resource_name --type ReleaseBundle --operation IN)

  res_types=( $buildinfo_res_name $filespec_res_name $releaseBundle_res_name )

  if [ "${#res_types[@]}" != 1 ]; then
    execute_command "echo Exactly one resource of type BuildInfo\|FileSpec\|ReleaseBundle is supported."
    execute_command "exit 1"
  fi

  # We put everything we want to upload to vms in a directory
  # We will create a tarball from all of it
  local tardir="$step_tmp_dir/deploy-artifacts"
  mkdir "$tardir"

  # Create a file with env vars to source on the target vms
  local vm_env_filename="$step_name-$run_id.env"
  local vm_env_file_path="$tardir/$vm_env_filename"

  vm_env_vars_len_config_val=$(find_step_configuration_value "vmEnvironmentVariables_len")
  if [ -n "$vm_env_vars_len_config_val" ];then
    for ((i=0; i<(vm_env_vars_len_config_val); i++)); do
      env_var=$(find_step_configuration_value "vmEnvironmentVariables_$i")
      echo export $(echo "$env_var") >> "$vm_env_file_path"
    done
  fi

  pushd "$tardir"
    if [ -n "$buildinfo_res_name" ]; then
      buildinfo_integration_alias=$(find_resource_variable "$buildinfo_res_name" integrationAlias)
      downloadBuildInfo "$buildinfo_res_name" "$buildinfo_integration_alias"
    elif [ -n "$filespec_res_name" ]; then
      # move the fileSpecs to tardir
      # no need to download because filespecs are automatically downloaded already.
      local filespec_res_path=$(eval echo "$"res_"$filespec_res_name"_resourcePath)
      execute_command "mv $filespec_res_path/* $tardir/"
    elif [ -n  "$releaseBundle_res_name" ]; then
      local rt_int_name=$(get_integration_name --type "Artifactory")
      if [ -z "$rt_int_name" ]; then
        execute_command "echo An Artifactory integration is required to use a ReleaseBundle resource with LinuxVMDeploy steps."
        execute_command "exit 1"
      fi
      downloadReleaseBundle "$rt_int_name" "$releaseBundle_res_name"
    fi
    # create tarball from everything in the tardir
    local tarball_name="$step_name-$run_id.tar.gz"
    execute_command "tar -czvf $step_tmp_dir/$tarball_name ."
  popd

  local target_dir="~/$step_name/$run_id"
  local target_dir_config_val=$(find_step_configuration_value "targetDirectory")
  if [ -n "$target_dir_config_val" ]; then
    target_dir=$target_dir_config_val
  fi

  local ssh_user_config_val=$(find_step_configuration_value "sshUser")
  local rollout_delay_config_val=$(find_step_configuration_value "rolloutDelaySecs")
  local fast_fail_config_val=$(find_step_configuration_value "fastFail")

  # export name of target cluster for use in commands defined in step configuration (e.g. deployCommand).
  export target_cluster_name=$(echo "$vm_cluster_names" | jq .[0] --raw-output)

  # Used for blueGreen strategy only. The cluster that was live at the beginning of the step.
  export current_cluster_name=""

  # For blueGreen, try to determine from step state which cluster to deploy to.
  # If we have nothing in state, just use the first one.
  local last_deployed_cluster_filename="last_deployed_cluster.txt"
  local last_deployed_cluster_filepath="$step_workspace_dir/$last_deployed_cluster_filename"
  if [ "$strategy" == "$BLUE_GREEN_STRATEGY" ]; then
    current_cluster_name=$(echo "$vm_cluster_names" | jq .[1] --raw-output)
    execute_command "restore_cache_files $last_deployed_cluster_filename $last_deployed_cluster_filepath"
    local cached_last_deployed_cluster=$(cat "$last_deployed_cluster_filepath")
    # If we have cached the name of the last deployed cluster, deploy to the other one.
    if [ -n "$cached_last_deployed_cluster" ] && [ "$cached_last_deployed_cluster" == "$target_cluster_name" ]; then
      current_cluster_name="$cached_last_deployed_cluster"
      target_cluster_name=$(echo "$vm_cluster_names" | jq .[1] --raw-output)
      execute_command "echo 'Previous deployment was to cluster: $cached_last_deployed_cluster. Deploying to cluster: $target_cluster_name'"
    fi
  fi

  local res_targets=$(eval echo "$"res_"$target_cluster_name"_targets)
  local vm_targets=( $(echo "$res_targets" | jq --raw-output '.[]') )

  if [ -n "$deploy_targets_override" ] && [ "$strategy" != "$BLUE_GREEN_STRATEGY" ]; then
    execute_command "echo 'Overriding vm deploy targets with: $deploy_targets_override'"
    IFS=,
    vm_targets=($deploy_targets_override)
    unset IFS
  fi

  deploy_failed=false
  on_failure="deploy_failed=true; break"
  # Don't exit on failed commands if fastFail is specified as false
  if [ -n "$fast_fail_config_val" ] && [ "$fast_fail_config_val" == false ]; then
    on_failure="deploy_failed=true; continue"
  fi

  # Upload artifact archive and env file to vms on target_cluster,
  # and then decompress the archive.
  for i in "${!vm_targets[@]}"
  do
    local vm_target=$(eval echo "${vm_targets[$i]}")

    # Wait between deploys if delay was specified
    if [ -n "$rollout_delay_config_val" ] && [ "$i" != 0 ]; then
      execute_command "sleep ${rollout_delay_config_val}s"
    fi
    local ssh_base_command="ssh $ssh_user_config_val@$vm_target -o StrictHostKeyChecking=no"
    local make_target_dir_command="$ssh_base_command \"mkdir -p $target_dir\""

    # Command to upload app tarball to vm
    local upload_command="scp $step_tmp_dir/$tarball_name $ssh_user_config_val@$vm_target:$target_dir"

    # Command to untar the uploaded tarball
    local untar="cd $target_dir/; tar -xf $tarball_name; rm -f $tarball_name;"

    # Command to decompress the uploaded artifacts so they can be used
    # in commands run with execute_on_target_cluster
    local untar_command="$ssh_base_command \"$untar\""

    if [ "$dry_run" == true ]; then
      local upload_command="echo 'dry_run is true. Skipping artifact upload.'"
      local untar_command="$ssh_base_command \"echo 'dry_run is true. Skipping artifact decompression.'\""
    fi
    execute_command "echo Creating target dir on vm."
    execute_command "$make_target_dir_command" || eval "$on_failure"
    execute_command "echo Uploading artifacts to vm."
    execute_command "$upload_command" || eval "$on_failure"
    execute_command "echo Decompressing artifacts archive on vm."
    execute_command "$untar_command" || eval "$on_failure"
  done

  exitIfDeployFailed

  deployed_vms=()
  if [ -n "$step_configuration_scripts_len" ]; then
    for i in $(seq 0 $(("$step_configuration_scripts_len" - 1)))
    do
      local script_context=$(find_step_configuration_value scripts_"$i"_context)
      local script_commands_json=$(eval echo "$"step_configuration_scripts_"$i"_commands)
      local script_commands=$(echo "$script_commands_json" | jq --raw-output '. | join("; ")')
      local script_name=$(find_step_configuration_value scripts_"$i"_name)
      if [ -z "$script_name" ]; then
        script_name="$DEFAULT_SCRIPT_NAME"
      fi
      current_command="$script_name"
      case $script_context in
        "$SCRIPT_CONTEXT_BUILD_NODE")
          execute_command "execute_on_build_node $script_name \"\$script_commands\""
          ;;
        "$SCRIPT_CONTEXT_CURRENT_CLUSTER")
          if [ "$strategy" != "$BLUE_GREEN_STRATEGY" ]; then
            execute_command "echo 'Execution of commands in context $SCRIPT_CONTEXT_CURRENT_CLUSTER is only supported for $BLUE_GREEN_STRATEGY strategy.'"
            execute_command "exit 1"
          fi
          execute_on_current_cluster "$script_name" "$script_commands"
          ;;
        "$SCRIPT_CONTEXT_TARGET_CLUSTER")
          execute_on_target_cluster "$script_name" "$script_commands"
          ;;
        *)
          execute_command "echo Unrecognized script context \'$script_context\' for script \'$script_name\'."
          execute_command "exit 1"
          ;;
      esac
      current_command=""
    done
  fi

  # Deploy was successful.

  # Handle success for blueGreen strategy.
  if [ "$strategy" == "$BLUE_GREEN_STRATEGY" ]; then
    # Save which cluster was deployed to in step state.
    execute_command "echo 'Saving last deployed cluster for blueGreen deployment: $target_cluster_name'"
    echo "$target_cluster_name" > "$last_deployed_cluster_filepath"
    execute_command "add_cache_files $last_deployed_cluster_filepath $last_deployed_cluster_filename"
  # Handle success for non-blueGreen strategies.
  elif [ "$dry_run" != true ]; then
    # Create rollback archive.
    for vm_target in "${deployed_vms[@]}"; do
      local ssh_base_command="ssh $ssh_user_config_val@$vm_target -o StrictHostKeyChecking=no"
      local rollback_dir="~/$step_name/rollback"
      # Command to copy artifacts into rollback dir.
      create_rollback_artifacts="$ssh_base_command \"mkdir -p $rollback_dir; rm -rf $rollback_dir/*; cp -r $target_dir/* $rollback_dir\""
      execute_command "echo 'Archiving successful deploy for rollback'"
      execute_command "$create_rollback_artifacts"
    done
  fi
}

DeployApplication
